package apitoken

import (
	"context"
	"crypto/rand"
	"errors"
	"fmt"
	"log"
	"net/http"

	"github.com/volatiletech/authboss/v3"
	"golang.org/x/crypto/bcrypt"
)

const (
	PageAPIToken              = "api_token"
	PageAPITokenConfirmDelete = "api_token_confirm_delete"
)

type Token interface {
	// GetPID gets the PID of the token owner
	GetPID() string
	// PutPID sets the PID of the token owner
	PutPID(string)

	GetName() string
	PutName(string)

	// GetID gets the ID of the token itself
	GetID() string
	PutID(string)

	GetSecret() string
	PutSecret(string)
}

type TokenServerStorer interface {
	// LoadToken loads the token by its ID
	LoadToken(ctx context.Context, id string) (Token, error)
	// SaveToken persists the token in the database, this should never
	// create a token and instead return ErrUserNotFound if the token
	// does not exist.
	SaveToken(ctx context.Context, token Token) error
	// NewToken creates a blank token, it is not yet persisted in the database
	// but is just for storing data
	NewToken(ctx context.Context) Token
	// CreateToken creates the token in storage, it should not overwrite a token
	// and should return ErrUserFound if it currently exists.
	CreateToken(ctx context.Context, token Token) error
	// DeleteToken deletes the token in storage
	DeleteToken(ctx context.Context, token Token) error
	ListTokens(ctx context.Context, pid string) ([]Token, error)
	// GetTokenName is used for including the token friendly name on a delete confirmation prompt
	GetTokenName(ctx context.Context, pid string, id string) (*string, error)
}

// EnsureTokenStorer makes sure the server storer supports api token operations
func EnsureTokenStorer(storer authboss.ServerStorer) TokenServerStorer {
	s, ok := storer.(TokenServerStorer)
	if !ok {
		panic("could not upgrade ServerStorer to TokenServerStorer, check your struct")
	}

	return s
}

// APIToken module.
type APIToken struct {
	*authboss.Authboss
}

func (r *APIToken) Setup() {
	authboss.RegisterModule("api_token", r)
	if err := r.Init(r.Authboss); err != nil {
		log.Println(err)
	}

}

func (r *APIToken) RequireTokenMiddleware(next http.Handler) http.Handler {
	f := func(w http.ResponseWriter, req *http.Request) {
		userID, err := r.ValidateToken(req)
		if err != nil {
			r.Authboss.RequestLogger(req).Error(err.Error())
			w.WriteHeader(http.StatusInternalServerError)
			return
		}
		user, err := r.Authboss.Storage.Server.Load(req.Context(), userID)
		if err == authboss.ErrUserNotFound {
			w.WriteHeader(http.StatusUnauthorized)
			return
		}
		req = req.WithContext(context.WithValue(req.Context(), authboss.CTXKeyUser, user))
		authboss.PutSession(w, authboss.SessionKey, userID)
		authboss.DelSession(w, authboss.SessionHalfAuthKey)

		next.ServeHTTP(w, req)
	}
	return http.HandlerFunc(f)
}

// Init the module.
func (r *APIToken) Init(ab *authboss.Authboss) (err error) {
	r.Authboss = ab

	if _, ok := ab.Config.Storage.Server.(TokenServerStorer); !ok {
		return errors.New("api_token module activated but storer could not be upgraded to TokenServerStorer")
	}

	ab.Config.Core.Router.Post("/api_token", ab.Config.Core.ErrorHandler.Wrap(r.Post))
	ab.Config.Core.Router.Delete("/api_token", ab.Config.Core.ErrorHandler.Wrap(r.Delete))
	// HTML forms don't support DELETE, so we need a post handler for the web client.
	ab.Config.Core.Router.Post("/api_token/delete", ab.Config.Core.ErrorHandler.Wrap(r.DeleteWeb))
	ab.Config.Core.Router.Get("/api_token/confirm_delete", ab.Config.Core.ErrorHandler.Wrap(r.ConfirmDeleteWeb))
	ab.Config.Core.Router.Get("/api_token", ab.Config.Core.ErrorHandler.Wrap(r.Get))

	return nil
}

// Post to the api_token page to create a token
func (r *APIToken) Post(w http.ResponseWriter, req *http.Request) error {
	logger := r.RequestLogger(req)

	validator, err := r.Authboss.Config.Core.BodyReader.Read(PageAPIToken, req)
	if err != nil {
		return err
	}

	values := MustHaveNameValues(validator)

	userID, err := r.Authboss.CurrentUserID(req)
	if err != nil || userID == "" {
		return r.Config.Core.Responder.Respond(w, req, http.StatusUnauthorized, PageAPIToken, authboss.NewHTMLData("status", "error", "error", "user not logged in"))
	}

	// Put values into newly created user for storage
	storer := EnsureTokenStorer(r.Config.Storage.Server)
	token := storer.NewToken(req.Context())

	id, secret, err := r.generateIDandSecret()
	if err != nil {
		return err
	}

	hashedSecret, err := bcrypt.GenerateFromPassword([]byte(secret), r.Config.Modules.BCryptCost)
	if err != nil {
		return err
	}

	token.PutPID(userID)
	token.PutName(values.GetName())
	token.PutID(id)
	token.PutSecret(string(hashedSecret))

	err = storer.CreateToken(req.Context(), token)
	switch {
	case err != nil:
		return err
	}

	logger.Infof("created api token for user %s", userID)
	data := authboss.HTMLData{
		"id":     id,
		"secret": secret,
	}
	if name := token.GetName(); name != "" {
		data["name"] = name
	}
	return r.Config.Core.Responder.Respond(w, req, http.StatusCreated, PageAPIToken, data)
}

func (r *APIToken) Delete(w http.ResponseWriter, req *http.Request) error {
	userID, err := r.Authboss.CurrentUserID(req)
	if err != nil || userID == "" {
		return r.Config.Core.Responder.Respond(w, req, http.StatusUnauthorized, PageAPIToken, authboss.NewHTMLData("status", "error", "error", "user not logged in"))
	}

	// Put values into newly created user for storage
	storer := EnsureTokenStorer(r.Config.Storage.Server)
	id := req.URL.Query().Get("id")
	if id == "" {
		return r.Config.Core.Responder.Respond(w, req, http.StatusOK, PageAPIToken, authboss.NewHTMLData("error", "no token id specified"))
	}
	token := storer.NewToken(req.Context())
	token.PutID(id)
	token.PutPID(userID)
	err = storer.DeleteToken(req.Context(), token)
	if err != nil {
		return err
	}
	return nil
}

func (r *APIToken) ConfirmDeleteWeb(w http.ResponseWriter, req *http.Request) error {
	userID, err := r.Authboss.CurrentUserID(req)
	if err != nil || userID == "" {
		return r.Config.Core.Responder.Respond(w, req, http.StatusUnauthorized, PageAPIToken, authboss.NewHTMLData("status", "error", "error", "user not logged in"))
	}

	id := req.URL.Query().Get("id")
	if id == "" {
		return r.Config.Core.Responder.Respond(w, req, http.StatusOK, PageAPIToken, authboss.NewHTMLData("error", "no token id specified"))
	}
	storer := EnsureTokenStorer(r.Config.Storage.Server)
	name, err := storer.GetTokenName(req.Context(), userID, id)
	if err != nil {
		return err
	}
	data := authboss.HTMLData{
		"id":   id,
		"name": name,
	}
	return r.Config.Core.Responder.Respond(w, req, http.StatusOK, PageAPITokenConfirmDelete, data)
}

func (r *APIToken) DeleteWeb(w http.ResponseWriter, req *http.Request) error {
	err := r.Delete(w, req)
	if err != nil {
		return err
	}
	return r.Config.Core.Redirector.Redirect(w, req, authboss.RedirectOptions{
		Success:      "Token deleted",
		RedirectPath: "/web/dashboard",
	})
}

func (r *APIToken) Get(w http.ResponseWriter, req *http.Request) error {
	userID, err := r.Authboss.CurrentUserID(req)
	if err != nil || userID == "" {
		return r.Config.Core.Responder.Respond(w, req, http.StatusUnauthorized, PageAPIToken, authboss.NewHTMLData("status", "error", "error", "user not logged in"))
	}
	// Put values into newly created user for storage
	storer := EnsureTokenStorer(r.Config.Storage.Server)
	tokens, err := storer.ListTokens(req.Context(), userID)
	if err != nil {
		return err
	}
	data := authboss.HTMLData{
		"tokens": tokens,
	}
	return r.Config.Core.Responder.Respond(w, req, http.StatusCreated, PageAPIToken, data)
}

func (r *APIToken) ValidateToken(req *http.Request) (string, error) {
	tokenID, tokenSecret, ok := req.BasicAuth()
	if !ok {
		return "", nil
	}

	// Put values into newly created user for storage
	storer := EnsureTokenStorer(r.Config.Storage.Server)
	token, err := storer.LoadToken(req.Context(), tokenID)
	if err != nil {
		return "", fmt.Errorf("load token: %w", err)
	}
	err = bcrypt.CompareHashAndPassword([]byte(token.GetSecret()), []byte(tokenSecret))
	if err != nil {
		return "", fmt.Errorf("validate token secret: %w", err)
	}
	return token.GetPID(), nil
}

func (r *APIToken) generateIDandSecret() (id string, secret string, err error) {
	rawID := make([]byte, 15)
	_, err = rand.Read(rawID)
	if err != nil {
		return "", "", fmt.Errorf("generate token: %w", err)
	}
	rawSecret := make([]byte, 18)
	_, err = rand.Read(rawSecret)
	if err != nil {
		return "", "", fmt.Errorf("generate secret: %w", err)
	}
	id = fmt.Sprintf("%x", rawID)
	secret = fmt.Sprintf("%x", rawSecret)
	return id, secret, nil
}

type NameValuer interface {
	authboss.Validator

	GetName() string
}

func MustHaveNameValues(v authboss.Validator) NameValuer {
	if n, ok := v.(NameValuer); ok {
		return n
	}

	panic(fmt.Sprintf("bodyreader returned a type that could not be upgraded to NameValuer: %T", v))
}
