package authentication

import (
	"encoding/base64"
	"fmt"
	"log"
	"net/http"
	"time"

	"git.codemonkeysoftware.net/mouseion/server/pkg/authentication/reader"
	"git.codemonkeysoftware.net/mouseion/server/pkg/email"
	"git.codemonkeysoftware.net/mouseion/server/pkg/views"

	"github.com/gorilla/sessions"
	abclientstate "github.com/volatiletech/authboss-clientstate"
	"github.com/volatiletech/authboss/v3"
	"github.com/volatiletech/authboss/v3/defaults"

	// "github.com/volatiletech/authboss/v3/otp/twofactor"
	// "github.com/volatiletech/authboss/v3/otp/twofactor/totp2fa"

	// authboss modules
	_ "git.codemonkeysoftware.net/mouseion/server/pkg/authentication/apitoken"
	_ "git.codemonkeysoftware.net/mouseion/server/pkg/authentication/confirm"
	_ "github.com/volatiletech/authboss/v3/auth"
	_ "github.com/volatiletech/authboss/v3/logout"
	_ "github.com/volatiletech/authboss/v3/register"
)

type Config struct {
	CookieStoreHashKey  string
	SessionStoreHashKey string
	SiteURL             string `config:"SITE_URL"`
	Development         bool
}

func Authboss(authStorer authboss.CreatingServerStorer, config Config, emailConfig email.Config) (*authboss.Authboss, error) {
	cookieStoreKey, err := base64.StdEncoding.DecodeString(config.CookieStoreHashKey)
	if err != nil {
		return nil, fmt.Errorf("authboss: cookie store key: %w", err)
	}
	sessionStoreKey, err := base64.StdEncoding.DecodeString(config.SessionStoreHashKey)
	if err != nil {
		log.Fatal("authboss: session store key:", err)
	}
	cookieStore := abclientstate.NewCookieStorer(cookieStoreKey, nil)
	sessionStore := abclientstate.NewSessionStorer("mouseion", sessionStoreKey, nil)

	cstore := sessionStore.Store.(*sessions.CookieStore)
	cstore.MaxAge(int((30 * 24 * time.Hour) / time.Second))

	if config.Development {
		log.Println("Auth cookie stores running in insecure mode (allows http)")

		cookieStore.Secure = false
		cstore.Options.Secure = false
	}

	ab := authboss.New()

	ab.Config.Paths.Mount = "/auth"
	ab.Config.Paths.ConfirmNotOK = "/account/confirmation_pending"
	ab.Config.Paths.ConfirmOK = "/auth/login?confirmed=true"
	ab.Config.Paths.AuthLoginOK = "/web/dashboard"
	ab.Config.Paths.RootURL = config.SiteURL
	ab.Config.Storage.Server = authStorer

	authTemplateRenderer := NewTemplateSetRenderer(views.WebTemplates(), "/pages/auth/")

	ab.Config.Core.ViewRenderer = NewHybridRenderer(authTemplateRenderer,
		defaults.JSONRenderer{})
	ab.Config.Core.MailRenderer = NewTemplateSetRenderer(views.EmailTemplates(), "")
	defaults.SetCore(&ab.Config, true, false)
	ab.Config.Core.BodyReader = reader.NewHTTPBodyReader(true, false)

	ab.Config.Core.ErrorHandler = defaults.NewErrorHandler(ab.Core.Logger)

	ab.Config.Mail.From = emailConfig.FromAddress
	ab.Config.Mail.FromName = emailConfig.FromName
	if emailConfig.SMTPHost != "" {
		ab.Config.Core.Mailer = email.NewMailer(emailConfig)
	} else {
		log.Println("SMTP host not set. Emails will be logged but not sent.")
	}

	ab.Config.Modules.RegisterPreserveFields = []string{"email"}
	ab.Config.Modules.ResponseOnUnauthed = authboss.RespondRedirect
	ab.Config.Modules.TOTP2FAIssuer = config.SiteURL
	ab.Config.Modules.LogoutMethod = http.MethodGet
	ab.Config.Modules.ConfirmMethod = http.MethodGet

	ab.Config.Storage.SessionState = sessionStore
	ab.Config.Storage.CookieState = cookieStore

	// // Set up 2fa
	// twofaRecovery := &twofactor.Recovery{Authboss: ab}
	// if err := twofaRecovery.Setup(); err != nil {
	// 	return nil, fmt.Errorf("authboss: %w", err)
	// }
	// totp := &totp2fa.TOTP{Authboss: ab}
	// if err := totp.Setup(); err != nil {
	// 	return nil, fmt.Errorf("authboss: %w", err)
	// }

	if err := ab.Init(); err != nil {
		return nil, fmt.Errorf("authboss: %w", err)
	}

	return ab, nil
}
