package sqlstore

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"time"

	"git.codemonkeysoftware.net/mouseion/server/pkg/authentication"
	"github.com/jmoiron/sqlx"
	"github.com/mattn/go-sqlite3"
	"github.com/volatiletech/authboss/v3"
)

var (
	ErrNoUsers = errors.New("no users in database")
)

type SQLStore struct {
	db *sqlx.DB
}

func New(db *sqlx.DB) *SQLStore {
	return &SQLStore{
		db: db,
	}
}

func (store *SQLStore) Load(ctx context.Context, email string) (authboss.User, error) {
	user := authentication.User{}
	err := store.db.GetContext(ctx, &user, "SELECT id, email, password, confirmed FROM auth_user WHERE email = ?",
		email,
	)
	if err != nil {
		if err == sql.ErrNoRows {
			return nil, authboss.ErrUserNotFound
		}
		return nil, fmt.Errorf("load user: %w", err)
	}
	return &user, nil
}

func (store *SQLStore) Save(ctx context.Context, user authboss.User) error {
	authUser, err := authentication.GetAuthUser(user)
	if err != nil {
		return fmt.Errorf("save: %w", err)
	}

	tx, err := store.db.BeginTxx(ctx, nil)
	if err != nil {
		return fmt.Errorf("begin tx: %w", err)
	}
	defer func() {
		_ = tx.Rollback()
	}()
	_, err = tx.ExecContext(ctx, `
UPDATE auth_user 
SET
  email = ?,
  confirmed = ?
WHERE id = ?`,
		authUser.Email,
		authUser.Confirmed,
		authUser.ID,
	)
	if err != nil {
		return fmt.Errorf("save user: %w", err)
	}

	if authUser.Password != "" {
		err = store.savePassword(ctx, tx, authUser)

		if err != nil {
			return fmt.Errorf("save password: %w", err)
		}
	}

	if authUser.ConfirmSelector == "" {
		err = store.deleteUserConfirmation(ctx, tx, authUser)
		if err != nil {
			return fmt.Errorf("delete user confirmation: %w", err)
		}
	} else {
		err = store.saveConfirmSelector(ctx, tx, authUser)
		if err != nil {
			return fmt.Errorf("save user confirmation: %w", err)
		}
	}
	err = tx.Commit()
	if err != nil {
		return fmt.Errorf("commit tx: %w", err)
	}
	return nil
}

func (store *SQLStore) New(ctx context.Context) authboss.User {
	return &authentication.User{}
}

func (store *SQLStore) Create(ctx context.Context, user authboss.User) error {
	authUser, err := authentication.GetAuthUser(user)
	if err != nil {
		return fmt.Errorf("create: %w", err)
	}

	tx, err := store.db.BeginTxx(ctx, nil)
	if err != nil {
		return fmt.Errorf("begin tx: %w", err)
	}
	defer func() {
		_ = tx.Rollback()
	}()
	err = tx.GetContext(ctx, &authUser.ID, `
INSERT INTO auth_user (email, password, confirmed)
VALUES (?,?,?)
RETURNING id`, authUser.Email, authUser.Password, authUser.Confirmed)
	if err != nil {
		if sqliteErr, ok := err.(sqlite3.Error); ok {
			if sqliteErr.ExtendedCode == sqlite3.ErrConstraintUnique {
				return authboss.ErrUserFound
			}
		}
		return fmt.Errorf("create: %w", err)
	}
	if authUser.ConfirmSelector == "" {
		err = store.deleteUserConfirmation(ctx, tx, authUser)
		if err != nil {
			return fmt.Errorf("delete user confirmation: %w", err)
		}
	} else {
		err = store.saveConfirmSelector(ctx, tx, authUser)
		if err != nil {
			return fmt.Errorf("save user confirmation: %w", err)
		}
	}
	err = tx.Commit()
	if err != nil {
		return fmt.Errorf("commit tx: %w", err)
	}
	return nil
}

func (store *SQLStore) UsersExist(ctx context.Context) error {
	var exist bool
	err := store.db.GetContext(ctx, &exist, "SELECT EXISTS(SELECT id FROM auth_user LIMIT 1)")
	if err != nil {
		return fmt.Errorf("users exist: %w", err)
	}
	if !exist {
		return ErrNoUsers
	}
	return nil
}

func (store *SQLStore) LoadByConfirmSelector(ctx context.Context, selector string) (authboss.ConfirmableUser, error) {
	var user authentication.User
	err := store.db.GetContext(ctx, &user, `
SELECT id, email, selector as confirm_selector, verifier as confirm_verifier
FROM auth_user JOIN user_confirmation ON auth_user.id = user_confirmation.user_id
WHERE selector = ?
AND user_confirmation.created_at > ?
`, selector, time.Now().Add(-2*time.Hour).UnixNano())
	if err != nil {
		if err == sql.ErrNoRows {
			return nil, authboss.ErrUserNotFound
		}
		return nil, fmt.Errorf("load user by confirm selector: %w", err)
	}
	return &user, nil
}

func (store *SQLStore) saveConfirmSelector(ctx context.Context, tx *sqlx.Tx, authUser *authentication.User) error {
	_, err := tx.ExecContext(ctx, `
INSERT INTO user_confirmation (user_id, selector, verifier,created_at)
VALUES (?,?,?,?)
ON CONFLICT (user_id)
  DO UPDATE
  SET selector = excluded.selector,
      verifier = excluded.verifier,
	  created_at = excluded.created_at`,
		authUser.ID,
		authUser.ConfirmSelector,
		authUser.ConfirmVerifier,
		time.Now().UnixNano(),
	)

	if err != nil {
		return fmt.Errorf("save confirmation: %w", err)
	}
	return nil
}

func (store *SQLStore) savePassword(ctx context.Context, tx *sqlx.Tx, authuser *authentication.User) error {
	_, err := tx.ExecContext(ctx, `
UPDATE auth_user 
SET password = ?
WHERE id = ?`,
		authuser.Password,
		authuser.ID)
	return err
}

func (store *SQLStore) deleteUserConfirmation(ctx context.Context, tx *sqlx.Tx, authUser *authentication.User) error {
	_, err := tx.ExecContext(ctx, `
DELETE FROM user_confirmation
WHERE user_id = ?`,
		authUser.ID)
	if err != nil {
		return fmt.Errorf("delete user confirmation: %w", err)
	}
	return nil
}
