package sqlstore

import (
	"context"
	"errors"
	"fmt"
	"time"

	"git.codemonkeysoftware.net/mouseion/server/pkg/authentication"
	"git.codemonkeysoftware.net/mouseion/server/pkg/authentication/apitoken"
)

func (store *SQLStore) SaveToken(ctx context.Context, token apitoken.Token) error {
	return nil
}

func (store *SQLStore) LoadToken(ctx context.Context, id string) (apitoken.Token, error) {
	token := authentication.APIToken{}
	err := store.db.GetContext(ctx, &token, `SELECT email as user_id, api_token.id, hashed_secret 
	FROM api_token JOIN auth_user ON api_token.user_id = auth_user.id
	WHERE api_token.id = $1`, id)
	return &token, err
}

func (store *SQLStore) NewToken(ctx context.Context) apitoken.Token {
	return &authentication.APIToken{}
}

func (store *SQLStore) CreateToken(ctx context.Context, token apitoken.Token) error {
	authToken, ok := token.(*authentication.APIToken)
	if !ok {
		return errors.New("token was not an authentication token")
	}

	tx, err := store.db.BeginTxx(ctx, nil)
	if err != nil {
		return fmt.Errorf("begin tx: %w", err)
	}
	defer func() {
		_ = tx.Rollback()
	}()

	_, err = tx.ExecContext(ctx, `INSERT INTO api_token (user_id, id, hashed_secret,created_at)
VALUES ((SELECT id FROM auth_user WHERE email = $1),$2,$3,$4)`,
		authToken.UserID, authToken.ID, authToken.Secret, time.Now().UnixNano())
	if err != nil {
		return fmt.Errorf("create token: %w", err)
	}
	if authToken.Name != nil {
		_, err = tx.ExecContext(ctx, `INSERT INTO api_token_name (id, name) VALUES ($1,$2)`, authToken.ID, *authToken.Name)
		if err != nil {
			return fmt.Errorf("create api token name: %w", err)
		}
	}

	err = tx.Commit()
	if err != nil {
		return fmt.Errorf("commit token: %w", err)
	}
	return nil
}

func (store *SQLStore) DeleteToken(ctx context.Context, token apitoken.Token) error {
	authToken, ok := token.(*authentication.APIToken)
	if !ok {
		return errors.New("token was not an authentication token")
	}
	_, err := store.db.ExecContext(ctx, `DELETE FROM api_token WHERE id = $1 and user_id = (SELECT id FROM auth_user WHERE email = $2)`, authToken.ID, authToken.UserID)
	return err
}

func (store *SQLStore) ListTokens(ctx context.Context, userID string) ([]apitoken.Token, error) {
	tokens := []*authentication.APIToken{}
	err := store.db.SelectContext(ctx, &tokens, `SELECT id, name FROM api_token LEFT JOIN api_token_name USING (id) WHERE user_id = (SELECT id FROM auth_user WHERE email = $1)`, userID)
	if err != nil {
		return nil, fmt.Errorf("list tokens: %w", err)
	}
	apiTokens := make([]apitoken.Token, len(tokens))
	for i, token := range tokens {
		apiTokens[i] = token
	}
	return apiTokens, nil
}

func (store *SQLStore) GetTokenName(ctx context.Context, userID string, id string) (*string, error) {
	var token authentication.APIToken
	err := store.db.GetContext(ctx, &token, `SELECT id, name FROM api_token LEFT JOIN api_token_name USING (id) WHERE id = $1 AND user_id = (SELECT id FROM auth_user WHERE email = $2)`, id, userID)
	if err != nil {
		return nil, fmt.Errorf("get token friendly name: %w", err)
	}
	return token.Name, nil
}
