package email

import (
	"context"
	"fmt"

	"github.com/volatiletech/authboss/v3"
	gomail "gopkg.in/gomail.v2"
)

type Config struct {
	FromAddress  string `config:"FROM_ADDRESS"`
	FromName     string `config:"FROM_NAME"`
	SMTPHost     string `config:"SMTP_HOST"`
	SMTPPort     int    `config:"SMTP_PORT"`
	SMTPUsername string `config:"SMTP_USERNAME"`
	SMTPPassword string `config:"SMTP_PASSWORD"`
}

type SMTPMailer struct {
	*gomail.Dialer
}

func NewMailer(config Config) *SMTPMailer {
	return &SMTPMailer{
		Dialer: gomail.NewDialer(config.SMTPHost, config.SMTPPort, config.SMTPUsername, config.SMTPPassword),
	}
}

type AddressNameMismatchError struct {
	addressCount, nameCount int
	field                   string
}

func (e *AddressNameMismatchError) Error() string {
	return fmt.Sprintf("address and name counts don't match for field %q: %d ≠ %d", e.field, e.addressCount, e.nameCount)
}

func SetMultiAddressHeader(msg *gomail.Message, field string, addresses []string, names []string) error {
	if len(names) != 0 && len(addresses) != len(names) {
		return &AddressNameMismatchError{
			addressCount: len(addresses),
			nameCount:    len(names),
			field:        field,
		}
	}
	header := make([]string, len(addresses))
	for i, address := range addresses {
		var name string
		if len(names) != 0 {
			name = names[i]
		}
		header[i] = msg.FormatAddress(address, name)
	}
	msg.SetHeader(field, header...)
	return nil
}

func (s SMTPMailer) Send(_ context.Context, email authboss.Email) error {
	msg := gomail.NewMessage()
	msg.SetAddressHeader("From", email.From, email.FromName)
	msg.SetAddressHeader("ReplyTo", email.ReplyTo, email.ReplyToName)
	err := SetMultiAddressHeader(msg, "To", email.To, email.ToNames)
	if err != nil {
		return err
	}
	err = SetMultiAddressHeader(msg, "Bcc", email.Bcc, email.BccNames)
	if err != nil {
		return err
	}
	err = SetMultiAddressHeader(msg, "Cc", email.Cc, email.CcNames)
	if err != nil {
		return err
	}
	msg.SetHeader("Subject", email.Subject)
	msg.AddAlternative("text/plain", email.TextBody)
	msg.AddAlternative("text/html", email.HTMLBody)

	return s.DialAndSend(msg)
}
