package entry

import (
	"html"
	"strings"
	"time"
	"unicode/utf8"
)

const (
	DefaultMaxTextLength = 1024
	DefaultMaxTagLength  = 100
	DefaultMaxTagCount   = 10
)

type LogEntry struct {
	ID        int       `json:"id" db:"id"`
	UserEmail string    `json:"-"`
	Timestamp time.Time `json:"timestamp" db:"timestamp"`
	Text      string    `json:"text" db:"text"`
	Tags      []string  `json:"tags"`
}

func (entry *LogEntry) TagsString() string {
	return strings.Join(entry.Tags, ", ")
}

func (entry *LogEntry) HTMLText() string {
	text := html.EscapeString(entry.Text)
	text = strings.ReplaceAll(text, "\n", "<br>")
	return text
}

// TruncateToCodePoint truncates a UTF-8 string to the beginning of the last
// code point before or at maxLength.
//
// If len(s) ≤ maxLength, the string is unchanged.
// Otherwise, the string is truncated immediately before any valid code point
// that crosses maxLength, or at maxLength if there is none.
//
// maxLength must be at least 4 bytes.
func TruncateToCodePoint(maxLength int, s string) string {
	if len(s) <= maxLength {
		return s
	}
	for i := 1; i <= 3; i++ {
		r, width := utf8.DecodeRuneInString(s[maxLength-i:])
		if r != utf8.RuneError && width > i {
			return s[:maxLength-i]
		}
	}
	return s[:maxLength]
}

type Limiter struct {
	MaxTextLength int
	MaxTagLength  int
	MaxTagCount   int
}

func DefaultLimiter() Limiter {
	return Limiter{
		MaxTextLength: DefaultMaxTextLength,
		MaxTagLength:  DefaultMaxTagLength,
		MaxTagCount:   DefaultMaxTagCount,
	}
}

// Limit modifies e to meet the requirements of l.
// It does not deduplicate e.Tags before truncating it; therefore, an entry
// with tags [a,a,a,a,a,a,a,a,a,a,b], limited to 10, will end up with only [a].
func (l Limiter) Limit(e *LogEntry) {
	e.Text = TruncateToCodePoint(l.MaxTextLength, e.Text)

	if len(e.Tags) > l.MaxTagCount {
		e.Tags = e.Tags[:l.MaxTagCount]
	}

	for i, tag := range e.Tags {
		e.Tags[i] = TruncateToCodePoint(l.MaxTagLength, tag)
	}
}
