package sqlstore_test

import (
	"context"
	"fmt"
	"io"
	"log"
	"os"
	"testing"
	"time"

	authsql "git.codemonkeysoftware.net/mouseion/server/pkg/authentication/sql_store"
	"git.codemonkeysoftware.net/mouseion/server/pkg/database/sqlite"
	"git.codemonkeysoftware.net/mouseion/server/pkg/entry"
	sqlstore "git.codemonkeysoftware.net/mouseion/server/pkg/entry/sql_store"
	"git.codemonkeysoftware.net/mouseion/server/pkg/testhelpers"
	"github.com/jmoiron/sqlx"
)

type Tester interface {
	Helper()
	Fatal(args ...interface{})
}

type testData struct {
	dbName    string
	db        *sqlx.DB
	Store     *sqlstore.SQLStore
	UserEmail string
}

func SetUp(t Tester) testData {
	t.Helper()

	var td testData
	td.dbName = fmt.Sprintf("%s_test.db", testhelpers.RandString(12))
	td.db = sqlite.Open(td.dbName)
	err := sqlite.DoMigrations(td.db.DB)
	if err != nil {
		td.TearDown()
		t.Fatal(err)
	}

	td.Store = sqlstore.New(td.db)
	authStore := authsql.New(td.db)
	user := authStore.New(context.Background())
	td.UserEmail = testhelpers.GenerateRandomEmail()
	user.PutPID(td.UserEmail)
	err = authStore.Create(context.Background(), user)
	if err != nil {
		td.TearDown()
		t.Fatal(err)
	}

	return td
}

func (td testData) TearDown() {
	_ = td.db.Close()
	_ = os.Remove(td.dbName)
}

func TestEntryLifecycle(t *testing.T) {
	testData := SetUp(t)
	defer testData.TearDown()

	entryDate := time.Date(2011, 11, 11, 11, 11, 11, 11, time.UTC)
	const goodTag = "good"

	logEntry := entry.LogEntry{
		UserEmail: testData.UserEmail,
		Timestamp: entryDate,
		Text:      "peppero day",
		Tags:      []string{"snacks", "dates", goodTag},
	}
	err := testData.Store.Save(context.Background(), &logEntry)
	if err != nil {
		t.Fatal(err)
	}

	badEntry1 := entry.LogEntry{
		UserEmail: testData.UserEmail,
		Timestamp: entryDate.Add(-time.Minute),
		Text:      "peppero day1",
		Tags:      []string{"snacks", "dates"},
	}
	err = testData.Store.Save(context.Background(), &badEntry1)
	if err != nil {
		t.Fatal(err)
	}

	badEntry2 := entry.LogEntry{
		UserEmail: testData.UserEmail,
		Timestamp: entryDate.Add(time.Minute),
		Text:      "peppero day2",
		Tags:      []string{"snacks", "dates"},
	}
	err = testData.Store.Save(context.Background(), &badEntry2)
	if err != nil {
		t.Fatal(err)
	}

	results, err := testData.Store.GetEntries(
		context.Background(),
		testData.UserEmail,
		entryDate.Add(-time.Second*10),
		entryDate.Add(time.Second*10),
		nil,
	)

	if err != nil {
		t.Fatal(err)
	}
	if len(results) != 1 {
		t.Fatalf("expected 1 result, got %d", len(results))
	}
	if results[0].Text != logEntry.Text {
		log.Fatalf("result text was %s", results[0].Text)
	}
	if results[0].Timestamp != logEntry.Timestamp {
		log.Fatalf("expected timestamp %s, got %s", logEntry.Timestamp, results[0].Timestamp)
	}

	// gets all entries with empty times
	results, err = testData.Store.GetEntries(
		context.Background(),
		testData.UserEmail,
		time.Time{},
		time.Time{},
		nil,
	)
	if err != nil {
		log.Fatal(err)
	}

	if len(results) != 3 {
		t.Fatalf("expected 3 results, got %d", len(results))
	}

	// filters by tag
	results, err = testData.Store.GetEntries(
		context.Background(),
		testData.UserEmail,
		time.Time{},
		time.Time{},
		[]string{goodTag},
	)
	if err != nil {
		log.Fatal(err)
	}

	if len(results) != 1 {
		t.Fatalf("expected 1 results, got %d", len(results))
	}
}

func TestSaveDeduplicatesTags(t *testing.T) {
	testData := SetUp(t)
	defer testData.TearDown()

	const repeatedTag = "stuff"
	const singleTag = "things"

	entryTime := time.Now().Add(-time.Hour)
	ent := entry.LogEntry{
		UserEmail: testData.UserEmail,
		Timestamp: entryTime,
		Text:      "hello",
		Tags:      []string{repeatedTag, repeatedTag, singleTag},
	}
	err := testData.Store.Save(context.Background(), &ent)
	if err != nil {
		t.Fatal(err)
	}

	results, err := testData.Store.GetEntries(context.Background(), testData.UserEmail, entryTime.Add(-time.Hour), time.Now(), nil)
	if err != nil {
		t.Fatal(err)
	}
	if len(results) != 1 {
		t.Fatalf("expected 1 results, got %d", len(results))
	}
	result := results[0]
	if len(result.Tags) != 2 {
		t.Fatalf("expected 2 tags, got %d", len(result.Tags))
	}

	expectTag(t, repeatedTag, result.Tags)
	expectTag(t, singleTag, result.Tags)
}

func expectTag(t *testing.T, tag string, tags []string) {
	for _, tag2 := range tags {
		if tag == tag2 {
			return
		}
	}
	t.Errorf("does not contain expected tag %s", tag)
}

func BenchmarkGetRange(b *testing.B) {
	originalWriter := log.Default().Writer()
	log.Default().SetOutput(io.Discard)
	defer func() {
		log.Default().SetOutput(originalWriter)
	}()
	b.Run("100 x 10", func(b *testing.B) {
		const (
			entryCount = 100
			tagCount   = 10
			userEmail  = "test@example.com"
		)
		testData := SetUp(b)
		defer testData.TearDown()

		for i := 0; i < entryCount; i++ {
			logEntry := entry.LogEntry{
				Timestamp: time.Now(),
				Text:      "logentry",
				Tags:      make([]string, 0, tagCount),
			}
			for n := 0; n < tagCount; n++ {
				logEntry.Tags = append(logEntry.Tags, fmt.Sprintf("tag %d", n))
			}
			err := testData.Store.Save(context.Background(), &logEntry)
			if err != nil {
				b.Fatal(err)
			}
		}
		b.ResetTimer()
		_, err := testData.Store.GetEntries(context.Background(), userEmail, time.Time{}, time.Time{}, nil)
		b.StopTimer()
		if err != nil {
			b.Fatal(err)
		}
	})
}
