package views

import (
	"embed"
	"errors"
	"io"
	"io/fs"
	"net/http"

	"github.com/CloudyKit/jet/v6"
	"github.com/CloudyKit/jet/v6/loaders/httpfs"
)

//go:embed templates/*
var templates embed.FS

// TemplateSet keeps consumers from messing up globals in our *jet.Set.
type TemplateSet interface {
	Get(templatePath string) (t *jet.Template, err error)
	// Execute is a convenience method equivalent to calling TemplateSet.Get and (*jet.Template).Execute.
	Execute(w io.Writer, templatePath string, variables jet.VarMap, data interface{}) error
}

type templateSet struct {
	set *jet.Set
}

func (r templateSet) Get(templatePath string) (*jet.Template, error) {
	if r.set == nil {
		return nil, errors.New("templateGetter not initialized")
	}
	return r.set.GetTemplate(templatePath)
}

func (r templateSet) Execute(w io.Writer, templatePath string, variables jet.VarMap, data interface{}) error {
	template, err := r.Get(templatePath)
	if err != nil {
		return err
	}
	return template.Execute(w, variables, data)
}

var webTemplates, emailTemplates templateSet

func loadJetSet(fsys fs.FS, dir string) (*jet.Set, error) {
	subdir, err := fs.Sub(fsys, dir)
	if err != nil {
		return nil, err
	}
	loader, err := httpfs.NewLoader(http.FS(subdir))
	if err != nil {
		return nil, err
	}
	return jet.NewSet(loader), nil
}

func init() {
	var err error
	webTemplates.set, err = loadJetSet(templates, "templates/web")
	if err != nil {
		panic(err)
	}
	emailTemplates.set, err = loadJetSet(templates, "templates/email")
	if err != nil {
		panic(err)
	}
}

func EmailTemplates() TemplateSet {
	return emailTemplates
}

func WebTemplates() TemplateSet {
	return webTemplates
}

// Map is a syntactical convenience for creating map[string]interface{} as template contexts.
type Map map[string]interface{}
