package webserver

import (
	"context"
	"encoding/json"
	"fmt"
	"io/fs"
	"log"
	"net/http"
	"strings"
	"time"

	"git.codemonkeysoftware.net/mouseion/server/pkg/authentication"
	"git.codemonkeysoftware.net/mouseion/server/pkg/authentication/confirm"
	"git.codemonkeysoftware.net/mouseion/server/pkg/entry"

	"github.com/go-chi/chi/v5"
	"github.com/go-chi/chi/v5/middleware"
	"github.com/volatiletech/authboss/v3"
)

type Server struct {
	entryStorer            EntryStorer
	authStorer             AuthStorer
	tokenService           TokenService
	ab                     *authboss.Authboss
	requireTokenMiddleware func(http.Handler) http.Handler
	templates              Templateer
	public                 fs.FS
}

func New(
	entryStorer EntryStorer,
	authStorer AuthStorer,
	ab *authboss.Authboss,
	tokenService TokenService,
	requireTokenMiddleware func(http.Handler) http.Handler,
	templates Templateer,
	public fs.FS,
) *Server {
	return &Server{
		entryStorer:            entryStorer,
		authStorer:             authStorer,
		ab:                     ab,
		tokenService:           tokenService,
		requireTokenMiddleware: requireTokenMiddleware,
		templates:              templates,
		public:                 public,
	}
}

func (server *Server) Start() {
	mux := chi.NewRouter()

	mux.Use(
		middleware.RequestID,
		middleware.RealIP,
		middleware.Logger,
		middleware.Recoverer,
		authentication.SetRequestType,
		server.ab.LoadClientStateMiddleware,
	)

	server.mountStatic(mux)
	mux.With(server.requireTokenMiddleware).Post("/ingest.json", server.jsonIngest)
	mux.With(server.requireLoggedIn, confirm.Middleware(server.ab)).Get("/client/get_entries", server.apiGetEntries)

	mux.With(server.requireLoggedIn).Route("/web", server.webRoutes)
	mux.Route("/account", server.accountRoutes)

	mux.Group(func(mux chi.Router) {
		mux.Use(func(next http.Handler) http.Handler {
			return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
				if r.URL.Path == "/auth/login" && r.URL.Query().Get("confirmed") == "true" {
					data, ok := r.Context().Value(authboss.CTXKeyData).(authboss.HTMLData)
					if !ok {
						data = make(authboss.HTMLData)
					}
					data["confirmed"] = struct{}{}
					r = r.WithContext(context.WithValue(r.Context(), authboss.CTXKeyData, data))
				}
				next.ServeHTTP(w, r)
			})
		})
		mux.Mount("/auth", http.StripPrefix("/auth", server.ab.Config.Core.Router))
	})

	mux.Get("/", func(rw http.ResponseWriter, r *http.Request) {
		err := server.templates.Execute(rw, "pages/index", nil, nil)
		if err != nil {
			log.Println(err)
		}
	})

	log.Println("listening :8069")
	log.Println(http.ListenAndServe(":8069", mux))
}

func (server *Server) jsonIngest(w http.ResponseWriter, req *http.Request) {
	if req.Method != http.MethodPost {
		w.WriteHeader(http.StatusNotFound)
		return
	}
	defer func() {
		if err := req.Body.Close(); err != nil {
			log.Println(err)
		}
	}()
	ent := &entry.LogEntry{}
	decoder := json.NewDecoder(req.Body)
	err := decoder.Decode(ent)
	if err != nil {
		log.Println(err)
	}
	user, err := server.ab.CurrentUser(req)
	if err != nil {
		log.Println(err)
	}

	ent.UserEmail = user.GetPID()

	entry.DefaultLimiter().Limit(ent)

	err = server.entryStorer.Save(req.Context(), ent)
	if err != nil {
		log.Println(err)
		fmt.Fprintln(w, err)
	}
}

const (
	queryStart = "start"
	queryEnd   = "end"
	queryTags  = "tags"
)

func (server *Server) apiGetEntries(w http.ResponseWriter, req *http.Request) {
	entries, err := server.searchEntries(req)
	if err != nil {
		log.Println(err)
		w.WriteHeader(http.StatusInternalServerError)
		fmt.Fprint(w, err)
	}

	enc := json.NewEncoder(w)
	w.WriteHeader(http.StatusOK)
	err = enc.Encode(entries)

	if err != nil {
		log.Println(err)
		return
	}
}

var parseFormats = []string{
	time.RFC3339,
	time.RFC3339Nano,
	"02 Jan 06 15:04",
	"2006-01-02T15:04:05",
	"2006-01-02T15:04",
	"2006-01-02",
}

func (server *Server) searchEntries(req *http.Request) ([]entry.LogEntry, error) {
	queryValues := req.URL.Query()
	var (
		start, end time.Time
		tags       []string
	)
	startQuery := queryValues.Get(queryStart)
	if startQuery != "" {
		var err error
		for i := 0; i < len(parseFormats) && err == nil && start.IsZero(); i++ {
			start, err = time.Parse(parseFormats[i], startQuery)
			if err != nil {
				switch err.(type) {
				case *time.ParseError:
					err = nil
				default:
					log.Println(err)
				}
			}
		}
	}
	log.Println(start)
	endQuery := queryValues.Get(queryEnd)
	if endQuery != "" {
		end, _ = time.Parse(time.RFC3339, endQuery)
	}
	tagsQuery := queryValues.Get(queryTags)
	if tagsQuery != "" {
		tags = strings.Split(tagsQuery, ",")
	}
	user, err := server.ab.CurrentUser(req)
	if err != nil {
		return nil, err
	}

	entries, err := server.entryStorer.GetEntries(req.Context(), user.GetPID(), start, end, tags)
	if err != nil {
		return nil, err
	}
	return entries, nil
}

func (server *Server) requireLoggedIn(next http.Handler) http.Handler {
	f := func(w http.ResponseWriter, r *http.Request) {
		_, err := server.ab.CurrentUser(r)
		if err == authboss.ErrUserNotFound {
			if r.Header.Get("Content-Type") == "application/json" {
				w.WriteHeader(http.StatusUnauthorized)
				return
			}
			w.Header().Set("Location", "/auth/login")
			w.WriteHeader(http.StatusUnauthorized)
			err := server.templates.Execute(w, "pages/must_log_in", nil, nil)
			if err != nil {
				log.Println(err)
			}
			return
		}
		next.ServeHTTP(w, r)
	}
	return http.HandlerFunc(f)
}
